unit WaterShed;


interface
uses Gr32,Classes,Windows,GeodesticTranformations,SysUtils;
type
TSortedMapItem = class
  Height :Integer;
  Positions:Array of TPoint;
  constructor Create(p_Height:Integer;p_pt:TPoint);
  procedure Add (p_pt:TPoint);

  private
  m_Size,m_Count:Integer;
  public
  property  Count:Integer read m_Count  ;
end;
TQueue =  class
  private m_List:TArrayOfPoint;
  m_Count,Size:Integer;


  public
    property Count:Integer read m_Count;
    constructor Create;
    procedure Add(p_pt:TPoint);
    function  First():TPoint;
    function Empty:Boolean;
end;
TSortedMap= class
   Items :TList;
   constructor Create;
   procedure InsertSorted(p_value:Integer;p_pt:TPoint);
end;
TSortedBitmap = class(TSortedMap)

  constructor Create(p_Bmp:TBitmap32) ;
  private




end;
TObjectCount = record
  Index:Integer;
  PixelsCount:Integer;
end;
TArrayOfObjectCount= Array of TObjectCount;
TWaterShed =  class
  private
  const mask =-2;
  const wshed =0;
  const init =-1;
  const inqueue = -3;
  private
  V_SortedOuttPutmap:TSortedMap;
  function Centroid(p_Points:Array of TPoint):TPoint;
  public
  procedure Apply(p_Srcbmp,p_Dstbmp:TBitmap32);overload;
  procedure Apply(p_Srcbmp,p_Mask,p_Dstbmp:TBitmap32);overload;
  procedure TestApplyMask(p_Srcbmp,p_Dstbmp:TBitmap32);
  function GetObjectCountArray:TArrayOfObjectCount;
  function GetObjectIndex(p_pt:TPoint):Integer;
  function GetObjectCount(p_Index:Integer):Integer;
  procedure DrawIndecesTo(p_DstBmp:TBitmap32);
  private
  function GetNeighBours(P_pt:TPoint;p_Width,p_Height:Integer):TArrayOfPoint;

end;
implementation
constructor TSortedMapItem.Create(p_Height: Integer; p_pt: TPoint);
begin
    SetLength(Positions,10);
    Height:=p_Height;
    m_Size :=10;
    m_Count :=1;
    Positions[0]:= p_pt;
end;
procedure TSortedMapItem.Add (p_pt:TPoint);
begin
   if m_Count=m_Size then  begin
      Inc(m_Size,10);
      SetLength(Positions,m_Size);
   end;

   Positions[m_Count]:= p_pt;
   inc(m_Count);

end;
procedure TWaterShed.TestApplyMask(p_Srcbmp,p_Dstbmp:TBitmap32);
var
v_MaskBmp:TBitmap32;
i,j:Integer;
begin
  v_MaskBmp:=TBitmap32.Create;
  v_MaskBmp.SetSize(p_Srcbmp.Width,p_Srcbmp.Height);
  for i := 0 to v_MaskBmp.Width-1 do
    for j :=0  to v_MaskBmp.Height-1 do begin
      if TColor32Entry(p_Srcbmp.Pixel[i,j]).R<90 then
        v_MaskBmp.Pixel[i,j] := clblack32
      else
        v_MaskBmp.Pixel[i,j]:= clWhite32;

    end;
// p_Dstbmp.SetSize(p_Srcbmp.Width,p_Srcbmp.Height);
  //v_MaskBmp.DrawTo(p_Dstbmp);
  //exit;


  Apply(p_Srcbmp,v_MaskBmp,p_Dstbmp);
end;
procedure TWaterShed.Apply(p_Srcbmp,p_Mask,p_Dstbmp:TBitmap32);
var
v_bmp,v_bmp2:TBitmap32;
begin
  v_bmp2:= Intersect(p_Srcbmp,p_Mask);
  v_bmp:=ReconstructionByErosion(p_Mask,v_bmp2);

  //v_bmp.DrawTo(p_Dstbmp);
  //exit;
  v_bmp2.Free;
  Apply(v_bmp,p_Dstbmp);
end;
procedure TWaterShed.Apply(p_Srcbmp: TBitmap32; p_Dstbmp: TBitmap32);
var
v_SortedBitmap:TSortedBitmap;
i,h,j,k:  Integer;

v_SortedMapItem:TSortedMapItem;
V_OutputImage:Array of Array of Integer;
v_neighbours:TArrayOfPoint;
v_Fifo:TQueue;
v_pt:TPoint;
v_flag:Boolean;
V_CurrentLabel:Integer;

begin
  V_CurrentLabel:=0;
  v_Fifo := TQueue.Create;
  //v_flag:= false;
  v_SortedBitmap := TSortedBitmap.Create(p_Srcbmp);
  SetLength(V_OutputImage,p_Srcbmp.Width);

  for i := 0 to p_Srcbmp.Width-1 do  begin
    SetLength(V_OutputImage[i],p_Srcbmp.Height);
    for j :=0  to  p_Srcbmp.Height-1 do
            V_OutputImage[i,j]:=init;
  end;

  for i := 0 to
    v_SortedBitmap.Items.Count-1
    do begin
      v_SortedMapItem :=TSortedMapItem( v_SortedBitmap.Items[i]);
      h:=v_SortedMapItem.Height;
      for j :=0  to v_SortedMapItem.Count-1  do begin
         V_OutputImage[v_SortedMapItem.Positions[j].X,v_SortedMapItem.Positions[
         j].Y] :=mask;
         v_neighbours := GetNeighBours(v_SortedMapItem.Positions[j],p_Srcbmp.Width,p_Srcbmp.Height);
         for k := low(v_neighbours)  to high(v_neighbours) do begin

           if (V_OutputImage[v_neighbours[k].X,v_neighbours[k].Y]>0) or
           (V_OutputImage[v_neighbours[k].X,v_neighbours[k].Y]= wshed) then
           begin
             V_OutputImage[v_SortedMapItem.Positions[j].X,v_SortedMapItem.Positions[
             j].Y] :=inqueue;

             v_Fifo.Add(v_SortedMapItem.Positions[j])

           end;
         end;
      end;
      while
      (not v_Fifo.Empty) do begin
         v_pt := v_Fifo.First;

         v_neighbours := GetNeighBours(v_pt,p_Srcbmp.Width,p_Srcbmp.Height);
         for k := low(v_neighbours)  to high(v_neighbours) do begin

           if (V_OutputImage[v_neighbours[k].X,v_neighbours[k].Y]>0)  then
           begin
             if  (V_OutputImage[v_pt.X,v_pt.Y]=inqueue) or ((V_OutputImage[v_pt.X,v_pt.Y]=wshed) and (v_flag=true))  then
               V_OutputImage[v_pt.X,v_pt.Y] := V_OutputImage[v_neighbours[k].X,v_neighbours[
                 k].Y]
             else if (V_OutputImage[v_pt.X,v_pt.Y]>0) and
             (V_OutputImage[v_pt.X,v_pt.Y]<>V_OutputImage[v_neighbours[k].X,
               v_neighbours[k].Y]) then begin
                 V_OutputImage[v_pt.X,v_pt.Y]:= wshed;
                 v_flag:= false;
               end
//               else begin
//                  Tlist.Create;
//               end;

           end else if(V_OutputImage[v_neighbours[k].X,v_neighbours[k].Y]=
             wshed)  then begin
                      if (V_OutputImage[v_pt.X,v_pt.Y] =inqueue) then begin
                          V_OutputImage[v_pt.X,v_pt.Y] := wshed;
                          v_flag := true;

                      end
                    end
              else if  V_OutputImage[v_neighbours[k].X,v_neighbours[k].Y]=mask
              then begin
                   V_OutputImage[v_neighbours[k].X,v_neighbours[k].Y]:= inqueue;
                   v_Fifo.Add(v_neighbours[k])
              end;

         end;

      end;

      for j :=0  to v_SortedMapItem.Count-1  do begin
         if  (V_OutputImage[v_SortedMapItem.Positions[j].X,v_SortedMapItem.Positions[
         j].Y] =mask) then
         begin
           inc(V_CurrentLabel);

           v_Fifo.Add(v_SortedMapItem.Positions[j]);
           V_OutputImage[v_SortedMapItem.Positions[j].X,v_SortedMapItem.Positions[
         j].Y] := V_CurrentLabel;
             while(not v_Fifo.Empty) do begin
               v_pt :=v_Fifo.First;
                v_neighbours := GetNeighBours(v_pt,p_Srcbmp.Width,p_Srcbmp.Height);
                for k := low(v_neighbours)  to high(v_neighbours) do begin
                   if V_OutputImage[v_neighbours[k].X,v_neighbours[k].Y]=mask
                   then begin
                       v_Fifo.add(v_neighbours[k]);
                       V_OutputImage[v_neighbours[k].X,v_neighbours[k].Y] :=
                         V_CurrentLabel;

                   end;
                end;

             end;

         end;

      end;


      {p_Dstbmp.SetSize(p_Srcbmp.Width,p_Srcbmp.Height);

      for k := 0 to p_Srcbmp.Width-1 do  begin
        for j :=0  to  p_Srcbmp.Height-1 do
           if V_OutputImage[k,j]=wshed then
                p_Dstbmp.PixelS[k,j]:= clRed32 else
           if  V_OutputImage[k,j]>0 then
               p_Dstbmp.PixelS[k,j] := Gray32(round(V_OutputImage[k,j]*255/V_CurrentLabel),255) ;

      end;}





  end;
    p_Dstbmp.SetSize(p_Srcbmp.Width,p_Srcbmp.Height);
   {for i := 0 to
    v_SortedBitmap.Items.Count-1
    do begin
      v_SortedMapItem :=TSortedMapItem( v_SortedBitmap.Items[i]);
      h:=v_SortedMapItem.Height;
      for j :=0  to v_SortedMapItem.Count-1  do begin
         p_Dstbmp.PixelS[v_SortedMapItem.Positions[j].X,v_SortedMapItem.Positions[
         j].Y]:=Gray32( h,255);
      end;
    end;}
    //p_Dstbmp.Clear;
   // V_SortedOuttPutBitmap:= TSortedBitmap.Create();
    V_SortedOuttPutmap:= TSortedMap.Create;

    for j :=0  to  p_Srcbmp.Height-1 do begin
        for k := 0 to p_Srcbmp.Width-1 do  begin
           if V_OutputImage[k,j]=wshed then
                p_Dstbmp.PixelS[k,j]:= clRed32;



           V_SortedOuttPutmap.InsertSorted(V_OutputImage[k,j],Point(k,j));
        end;

    end;

    DrawIndecesTo(p_Dstbmp);
    {{$apptype Console}


    {for i := 0 to
    V_SortedOuttPutmap.Items.Count-1
    do begin
      v_SortedMapItem :=TSortedMapItem( V_SortedOuttPutmap.Items[i]);
      Write(v_SortedMapItem.Height);
      Write(':');
      WriteLn(v_SortedMapItem.Count);
//      for j :=0  to v_SortedMapItem.Count-1  do begin
//         WriteLn('('+ IntToStr(v_SortedMapItem.Positions[j].X)+','+IntToStr( v_SortedMapItem.Positions[j].Y)+')');
//      end;
    end;


    Writeln('number of objects ='+ IntToStr(V_CurrentLabel));
    }
     //p_Dstbmp.SetSize(p_Srcbmp.Width,p_Srcbmp.Height);





end;
function TWaterShed.GetObjectCountArray:TArrayOfObjectCount;
var
i:Integer;
begin
     SetLength(Result,V_SortedOuttPutmap.Items.Count);
     for i := 0 to
    V_SortedOuttPutmap.Items.Count-1
    do begin
        Result[i].Index:=TSortedMapItem( V_SortedOuttPutmap.Items[i]).Height;
        Result[i].PixelsCount:=TSortedMapItem( V_SortedOuttPutmap.Items[i]).Count;
    end;
end;
procedure TWatershed.DrawIndecesTo(p_DstBmp:TBitmap32);
var
i:Integer;
v_SortedMapItem:TSortedMapItem;
p_pt:TPoint;
begin
   for i := 0 to
    V_SortedOuttPutmap.Items.Count-1
    do begin
      v_SortedMapItem :=TSortedMapItem( V_SortedOuttPutmap.Items[i]);
      p_pt:=Centroid(v_SortedMapItem.Positions);
      p_DstBmp.RenderText(p_pt.X,p_pt.Y,IntToStr( v_SortedMapItem.Height),
        0,clRed32);
//      for j :=0  to v_SortedMapItem.Count-1  do begin
//         WriteLn('('+ IntToStr(v_SortedMapItem.Positions[j].X)+','+IntToStr( v_SortedMapItem.Positions[j].Y)+')');
//      end;
    end; 
end;
function TWatershed.Centroid(p_Points:Array of TPoint):TPoint;
var
v_x,v_y:Cardinal;
i:Integer;
begin
{v_x:=0;
v_y:=0;
for i:=low(p_Points)  to high(p_Points) do begin
       v_x:= v_x+ p_Points[i].X;
       v_y:= v_y+ p_Points[i].Y;
end;
Result.X:= Round(v_x/(high(p_Points)-low(p_Points)+1));
Result.Y:= Round(v_y/(high(p_Points)-low(p_Points)+1));}
Result:=p_Points[Round((high(p_Points)-low(p_Points)+1)/2)]
;
if Result.X-10>0 then

Result.X:= Result.X-10 ;

if Result.Y-10>0 then
Result.Y:= Result.Y-10 ;
//Result:=p_Points[0];
end;
function TWaterShed.GetObjectIndex(p_pt:TPoint):Integer;
var
i,j:Integer;
v_SortedMapItem:TSortedMapItem;
begin
    for i := 0 to
    V_SortedOuttPutmap.Items.Count-1
    do begin
      v_SortedMapItem :=TSortedMapItem( V_SortedOuttPutmap.Items[i]);
      for j :=0  to v_SortedMapItem.Count-1  do begin
         if (v_SortedMapItem.Positions[j].X = p_pt.X)and(v_SortedMapItem.Positions[j].Y = p_pt.Y) then begin
          Result:= v_SortedMapItem.Height;
          exit;
         end;
      end;
    end;
end;
function TWaterShed.GetObjectCount(p_Index:Integer):Integer;
var
i,j:Integer;
v_SortedMapItem:TSortedMapItem;
begin
    for i := 0 to
    V_SortedOuttPutmap.Items.Count-1
    do begin
      v_SortedMapItem :=TSortedMapItem( V_SortedOuttPutmap.Items[i]);
      for j :=0  to v_SortedMapItem.Count-1  do begin
         if (v_SortedMapItem.Height=p_Index) then begin
          Result:= v_SortedMapItem.Count;
          exit;
         end;
      end;
    end;

end;
constructor TSortedMap.Create;
begin
  Items := TList.Create;
end;
procedure TSortedMap.InsertSorted(p_value:Integer;p_pt:TPoint);
var
i:Integer;
p_itemi:TSortedMapItem;
p_lastitem:TSortedMapItem;

begin

  for i :=0  to Items.Count-1  do begin

     p_itemi:=Items[i];
     //if ((p_itemi.Height - p_value)>=-2) and((p_itemi.Height - p_value)<=2) then begin
     if (p_itemi.Height = p_value) then begin
         p_itemi.Add(p_pt);
       exit;
     end;
     if (i=0) and  (p_value<p_itemi.Height) then begin
       Items.Insert(0,TSortedMapItem.Create(p_value,p_pt));
       exit;
     end;
     if (i=Items.Count-1) and  (p_value>p_itemi.Height) then begin
       Items.add(TSortedMapItem.Create(p_value,p_pt));
       exit;
     end;
     if {(p_value>p_lastitem.Height) and  }(p_value<p_itemi.Height) then begin
       Items.Insert(i,TSortedMapItem.Create(p_value,p_pt));
       exit;
     end;
     p_lastitem:=  p_itemi;


  end;
  Items.add(TSortedMapItem.Create(p_value,p_pt));
end;
constructor TSortedBitmap.Create(p_Bmp:TBitmap32) ;
var
i,j:Integer;
clrentry:TColor32Entry;

begin
  Items := TList.Create;


  for  i:=0  to p_Bmp.Width-1  do
    for j :=0  to p_Bmp.Height-1 do begin
      clrentry:= TColor32Entry(p_Bmp.Pixel[i,j]) ;
      {if  clrentry.R>=128 then
        InsertSorted( 255,Point(i,j))
      else
        InsertSorted( 0,Point(i,j));
      }
      InsertSorted( clrentry.R,Point(i,j));
      {if Items.Count>5 then begin
         Items[0]:= Items[0];
      end;}


    end;
end;
function TWatershed.GetNeighBours(P_pt:TPoint;p_Width,p_Height:Integer):TArrayOfPoint;
var
v_size,i,j,k:Integer;
begin
  v_size:=8;
  SetLength(Result,8);
  k:=0;
  for i := P_pt.X-1  to P_pt.X+1 do
    for j :=P_pt.Y-1  to P_pt.Y+1 do begin
      if (i=P_pt.X) and(j=P_pt.Y) then
        continue;
      if ((i>=0)and(i<p_Width)) and((j>=0)and(j<p_Height)) then begin
          Result[k]  := Point(i,j);
          inc(k);
      end else
        dec(v_size);
    end;

  SetLength(Result,v_size) ;

end;
procedure TQueue.Add(p_pt:TPoint);
begin
  if m_Count=Size then begin
    Inc(size,10);
    SetLength(m_List,Size);
  end;

  m_List[Count] :=p_pt ;
  inc(m_Count);

end;
constructor TQueue.Create;
begin
  Setlength(m_List,10);
  m_Count:=0;
  Size:=10;
end;

 function  TQueue.First():TPoint;
 var
 i:Integer;
 begin
  Result := m_List[0];
  for i := 1 to Count-1 do  begin
       m_List[i-1]:= m_List[i] ;
  end;
  dec(m_Count);
  if Size-m_Count>10 then begin
    dec(Size,10);
    SetLength(m_List,Size);
  end;


 end;
 function TQueue.Empty:Boolean;
 begin
      Result :=Count=0;
 end;
 end.

